/*
 * chattr.c		- Change file attributes on an ext2 file system
 *
 * Copyright (C) 1993, 1994  Remy Card <card@masi.ibp.fr>
 *                           Laboratoire MASI, Institut Blaise Pascal
 *                           Universite Pierre et Marie Curie (Paris VI)
 *
 * This file can be redistributed under the terms of the GNU General
 * Public License
 */

/*
 * History:
 * 93/10/30	- Creation
 * 93/11/13	- Replace stat() calls by lstat() to avoid loops
 * 94/02/27	- Integrated in Ted's distribution
 *
 * 95/02/17	- Completly rewriten to enable control over compression
 *		  parameters. Make sure we don't close the file after setting
 *		  the compress flag, and before setting the cluster size and
 *		  compression algorithm. Uses less space in the stack when
 *		  recursive.
 *
 *		  	Antoine de Maricourt.
 */

#include <dirent.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/stat.h>

#include <linux/ext2_fs.h>

const char *prog_name = "chattr";

int add = 0;
int rem = 0;
int set = 0;
int set_version  = 0;
unsigned long cluster_size = 0;
unsigned long compr_meth   = -1;

unsigned long version;

int recursive = 0;
int verbose   = 0;

unsigned long af = 0;
unsigned long rf = 0;
unsigned long sf = 0;

/*
 *	Do it for every file in the directory.
 */

static void chattr (char *path, int len);

static void chattr_dir (char *path, int len)
{
  DIR *dir = opendir (path);

  if (dir == NULL)
    return;

  else {
    struct dirent *de;

    path[len++] = '/';
    path[len]   = '\0';

    while ((de = readdir (dir)) != NULL)
      if (de->d_name[0] != '.' || (de->d_name[1] != '\0' && (de->d_name[1] != '.' || de->d_name[2] != '\0')))
	{ strcpy (path + len, de->d_name); chattr (path, len + de->d_reclen); }

    closedir (dir);
  }
}

/*
 *	Do it for a single file, and go deeper if recursive.
 */

static void chattr (char *path, int len)
{
  int fd;
  struct stat st;

  if (lstat (path, &st) != -1 && (S_ISDIR (st.st_mode) || S_ISREG (st.st_mode))) {
    if ((fd = open (path, O_RDONLY)) == -1)
      fprintf (stderr, "%s: %s: %s\n", prog_name, path, strerror (errno));

    else {
      unsigned long o_flags;
      
      /*
       *	First, get the old flags.
       */

      if (ioctl (fd, EXT2_IOC_GETFLAGS, &o_flags) == -1)
	fprintf (stderr, "%s: %s: can't get flags (%s)\n", prog_name, path, strerror (errno));

      else {
	unsigned long n_flags = o_flags;

	if (set)
	  n_flags = sf;

	else {
	  if (rem) n_flags &= ~rf;
	  if (add) n_flags |=  af;
	}
	
#if 0
	/* TODO: faire uniquement si ca a un sens et etre plus subtil ... */
	if (n_flag & EXT2_DISCOMPR_FL && ! (o_flags ...))
#endif
	  
	  if (ioctl (fd, EXT2_IOC_SETFLAGS, &n_flags) == -1)
	    fprintf (stderr, "%s: can't set flags (%s)\n", path, strerror (errno));
      }
  
      if (set_version && ioctl (fd, EXT2_IOC_SETVERSION, &version) == -1)
	fprintf (stderr, "%s: can't set version (%s)\n", path, strerror (errno));

      if (cluster_size && ioctl (fd, EXT2_IOC_SETCLUSTERSIZE, &cluster_size) == -1)
	fprintf (stderr, "%s: can't set cluster size (%s)\n", path, strerror (errno));
    
      if (compr_meth != -1 && ioctl (fd, EXT2_IOC_SETCOMPRMETHOD, &compr_meth) == -1)
	fprintf (stderr, "%s: can't set compression method (%s)\n", path, strerror (errno));

      close (fd);

      /*
       *	Do subdirectories.
       */

      if (recursive && S_ISDIR (st.st_mode)) 
	chattr_dir (path, len);
    }
  }
}

/*
 *	Display some help.
 */

static void usage ()
{
  fprintf (stderr, 
	   "usage: %s [-R] [-+=csuS] [-b size] [-v version] [-a name] files ...\n"
	   "  -R         recursive mode\n"
	   "  +-=c       set/reset compress flag\n"
	   "  +-=S       set/reset sync flag\n"
#if 0
	   "  +-=X       enable or disable compression\n"
#endif
	   "  +-=u       set/reset undelete flag\n"
	   "  +-=s       set/reset secure rm flag\n"
	   "  -a name    set compression algorithm\n"
	   "  -h         show this help message\n"
	   "  -b size    set cluster size\n"
	   "  -v number  set file version\n",
	   prog_name);

  exit (1);
}

/*
 *	Some predefined algorithm names.
 */

#define NA	11

static struct { char *name; int id; } table[11] = {
  { "lzv"	, EXT2_LZV_ID	 	},
  { "lzrw3a"	, EXT2_LZRW3A_ID	},
  { "defl.1"	, EXT2_DEFLATE_ID + 1	},
  { "defl.2"	, EXT2_DEFLATE_ID + 2	},
  { "defl.3"	, EXT2_DEFLATE_ID + 3	},
  { "defl.4"	, EXT2_DEFLATE_ID + 4	},
  { "defl.5"	, EXT2_DEFLATE_ID + 5	},
  { "defl.6"	, EXT2_DEFLATE_ID + 6	},
  { "defl.7"	, EXT2_DEFLATE_ID + 7	},
  { "defl.8"	, EXT2_DEFLATE_ID + 8	},
  { "defl.9"	, EXT2_DEFLATE_ID + 9	}
};

/*
 *	Parse arguments.
 */

static int decode_arg (int *i, int argc, char **argv)
{
  char *p;
  char *tmp;
  int   n = *i, x;

  switch (argv[n][0]) {

  case '-':
    for (p = &argv[n][1]; *p; p++)
      switch (*p) {

      case 'b' : 
	cluster_size = strtol (argv[++(*i)], &tmp, 0); if (*tmp) usage ();

	if (*tmp || (cluster_size !=  4 && cluster_size !=  8 &&
		     cluster_size != 16 && cluster_size != 32))
	  { fprintf (stderr, "%s: invalid cluster size: %s\n", prog_name, argv[*i]); exit (1); }

	break;

      case 'a' :
	for (x = 0, ++(*i); x < NA; x++)
	  if (strcmp (argv[*i], table[x].name) == 0) break;

	if (x == NA)
	  { fprintf (stderr, "%s: unknown algorithm: %s\n", prog_name, argv[*i]); exit (1); }

	compr_meth = table[x].id;
	break;

      case 'R' : recursive = 1;	break;
      case 'V' : verbose   = 1;	break; /* Unimplemented. Is it really usefull ? */
		
      case 'S' : rf |= EXT2_SYNC_FL;	 rem = 1; break;
      case 'c' : rf |= EXT2_COMPR_FL;	 rem = 1; break;
      case 's' : rf |= EXT2_SECRM_FL;	 rem = 1; break;
      case 'u' : rf |= EXT2_UNRM_FL;	 rem = 1; break;
#if 0
      case 'X' : rf |= EXT2_DISCOMPR_FL; rem = 1; break;
#endif

      case 'v' :
	version = strtol (argv[++(*i)], &tmp, 0);

	if (*tmp)
	  { fprintf (stderr, "%s: bad version: %s\n", prog_name, argv[*i]); exit (1); }

	set_version = 1;
	break;
	

      case 'h' : usage (); exit (0);

      default:
	fprintf (stderr, "%s: unrecognized argument: %c\n", prog_name, *p); usage ();
      }
    break;

  case '+':
    add = 1;

    for (p = &argv[n][1]; *p; p++)
      switch (*p) {
	case 'S' : af |= EXT2_SYNC_FL;	break;
	case 'c' : af |= EXT2_COMPR_FL;	break;
	case 's' : af |= EXT2_SECRM_FL;	break;
	case 'u' : af |= EXT2_UNRM_FL;	break;

	default  : usage ();
	}
    break;

  case '=' :
    set = 1;
	
    for (p = &argv[n][1]; *p; p++)
      switch (*p) {
      case 'S' : sf |= EXT2_SYNC_FL;	break;
      case 'c' : sf |= EXT2_COMPR_FL;	break;
      case 's' : sf |= EXT2_SECRM_FL;	break;
      case 'u' : sf |= EXT2_UNRM_FL;	break;

      default  : usage ();
      }
    break;

  default : return EOF;
  }
	
  return 1;
}

void main (int argc, char *argv[])
{
  int i;

  if (argc && *argv)
    prog_name = *argv;

  for (i = 1; i < argc; i++)
    if (decode_arg (&i, argc, argv) == EOF)
      break;

  if (i >= argc)
    usage ();

  if (set && (add || rem))
    { fprintf (stderr, "%s: = is incompatible with - and +\n", prog_name); exit (1); }

  while (i < argc)
    { chattr (argv[i], strlen (argv[i])); i++; }
}
