/*
 * lsattr.c		- List file attributes on an ext2 file system
 *
 * Copyright (C) 1993, 1994  Remy Card <card@masi.ibp.fr>
 *                           Laboratoire MASI, Institut Blaise Pascal
 *                           Universite Pierre et Marie Curie (Paris VI)
 *
 * This file can be redistributed under the terms of the GNU General
 * Public License
 */

/*
 * History:
 * 93/10/30	- Creation
 * 93/11/13	- Replace stat() calls by lstat() to avoid loops
 * 94/02/27	- Integrated in Ted's distribution
 *
 * 95/02/17	- Completly rewriten to display compression information
 *		  and to behave much more like ls. Useless calls to
 *		  intermediate routines that just call ioctl() were
 *		  replaced by direct calls to ioctl(). Uses less space
 *		  in the stack when recursive listing. Alphabetic listing
 *		  thanks to scandir().
 *
 *		  	Antoine de Maricourt.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/stat.h>

#include <linux/ext2_fs.h>

const char *prog_name = "lsattr";

int all = 0;
int d_opt = 0;
int l_opt = 0;
int recursive = 0;
int v_opt = 0;

/*
 *	Display attributes for one file.
 */

/* Pourquoi y-a-t-il un double / dans le nom des fichier en cas d'erreur ???
   'lsattr double' car on a en fait ecrit 'lsattr double/' ...! */

static void lsattr (char *path, char *name)
{
  int fd;

  if ((fd = open (path, O_RDONLY)) == -1)
    fprintf (stderr, "%s: %s: %s\n", prog_name, path, strerror (errno));

  else {
    unsigned long flags;
    unsigned long version;
    unsigned long size;
    unsigned long method	= -1;
    char          buf[5];

    if (ioctl (fd, EXT2_IOC_GETFLAGS      , &flags)   == -1 ||
	ioctl (fd, EXT2_IOC_GETVERSION    , &version) == -1 ||
	ioctl (fd, EXT2_IOC_GETCLUSTERSIZE, &size)    == -1 ||
	ioctl (fd, EXT2_IOC_GETCOMPRMETHOD, &method)  == -1)
      fprintf (stderr, "%s: %s: %s\n", prog_name, path, strerror (errno));

    else {

      /*
       *	Version
       */

      if (v_opt)
	printf ("%5lu ", version);

      /*
       *	Flags
       */

      buf[0] = (flags & EXT2_SYNC_FL)  ? 'S' : '-';
      buf[1] = (flags & EXT2_ECOMPR_FL ? 'X' : (flags & EXT2_COMPR_FL ? 'c' : '-'));
      buf[2] = (flags & EXT2_SECRM_FL) ? 's' : '-';
      buf[3] = (flags & EXT2_UNRM_FL)  ? 'u' : '-';
      buf[4] = '\0';

      fputs (buf, stdout);

      /*
       *	Cluster size
       */
	
      if ((flags & EXT2_COMPR_FL) || (flags & EXT2_COMPRBLK_FL)) 
	printf (" %2lu", size); else printf ("  -");

      /*
       *	Compression method
       */

      if (flags & EXT2_COMPR_FL) 
	switch (method) {
	case EXT2_LZV_ID	: printf (" lzv    "); break;
	case EXT2_LZRW3A_ID	: printf (" lzrw3a "); break;
	default:
	  printf (" ?      "); 
	}

      else printf (" -      ");

      printf (" %s\n", name);
    }

    close (fd);
  }
}

/*
 *	Used by scandir to select file names.
 */

static int select_dir (const struct dirent *d)
{
  return (d->d_name[0] != '.' ||
	  (all && d->d_name[1] != '\0' && (d->d_name[1] != '.' || d->d_name[2] != '\0')));
}

/*
 *	Show attributes for every files in a directory,
 *	and go deeper if recursive is true. Files are
 *	sorted in alphabetic order.
 */

static void lsattr_dir (char *path, int len, int echo)
{
  struct stat st;

  if (len > 0 && lstat (path, &st) == -1)
    { fprintf (stderr, "%s: %s: %s\n", prog_name, path, strerror (errno)); return; }

  if (len == 0 || S_ISDIR (st.st_mode)) {
    struct dirent **lst;
    int i, count;

    if (echo) printf ("\n%s:\n", path);

    /*
     *	Scan the directory and sort files in alphabetic order.
     */

    if ((count = scandir (len == 0 ? "." : path, &lst, select_dir, alphasort)) < 0)
      { fprintf (stderr, "%s: %s: %s\n", prog_name, path, strerror (errno)); return; }

    /*
     *	Do files, then subdirectories
     */

    if (len > 0)
      { path[len++] = '/'; path[len] = '\0'; }

    for (i = 0; i < count; i++)
      { strcpy (path + len, lst[i]->d_name); lsattr (path, lst[i]->d_name); }

    if (recursive)
      for (i = 0; i < count; i++)
	{ strcpy (path + len, lst[i]->d_name); lsattr_dir (path, len + lst[i]->d_reclen, 1); }

    /*
     *	Free the list
     */

    for (i = 0; i < count; i++)
      free (lst[i]);

    free (lst);
  }
}

/*
 *	Main routine
 */

static void volatile usage (void)
{
  fprintf (stderr, 
	   "usage: %s [-Radvh] [files ...]\n"
	   "  -R         recursive mode\n"
	   "  -a         show files whose name starts with '.'\n"
	   "  -d         directories behave like normal files\n"
	   "  -h         show this help message\n"
	   "  -v         display version number\n",
	   prog_name);

  exit (1);
}

void main (int argc, char ** argv)
{
  char path[MAXPATHLEN];
  char c;
  int i;

  if (argc && *argv)
    prog_name = *argv;

  while ((c = getopt (argc, argv, "Radhv")) != EOF)
    switch (c) {
    case 'R' : recursive = 1;	break;
    case 'a' : all = 1;		break;
    case 'd' : d_opt = 1;	break;
    case 'l' : l_opt = 1;	break;
    case 'v' : v_opt = 1;	break;
    case 'h' :
    default  : usage ();
    }

  path[0] = '\0';

  if (optind > argc - 1)
    { if (! d_opt) lsattr_dir (path, 0, 0); else lsattr (".", "."); }

  /* TODO: faire en sorte que 'lsattr -R .' ne prefixe pas avec './' */

  else for (i = optind; i < argc; i++) {
    if (! d_opt) {
      struct stat st;

      if (lstat (argv[i], &st) != -1)
	if (S_ISDIR (st.st_mode)) 
	  { strcpy (path, argv[i]); lsattr_dir (path, strlen (argv[i]), 0); }
	else lsattr (argv[i], argv[i]);
    
      else fprintf (stderr, "%s: %s: %s\n", prog_name, argv[i], strerror (errno));
    }

    else lsattr (argv[i], argv[i]);
  }
}
